import numpy as np
from .base import Operation, Property
from .point import LikePoint, Point
from typing import Sequence, Tuple, Union

class PointCloud(Point):
    def __init__(self, points: Sequence[LikePoint]) -> None: ...
    def transform(self, operation: Operation, **kwargs) -> Property: ...
LikePointCloud = Union[PointCloud, Sequence[LikePoint]]

class Corners(PointCloud):
    def get(self) -> np.ndarray: ...
    @staticmethod
    def product(shape: Tuple) -> Corners: ...



__all__ = (
    "Operation",
    "Property",
    "Constant",
    "Source",
    "Convert",
    "Iter",
    "RandomUniform",
    "RandomChoice",
    "Transformer",
)

from abc import ABC, abstractmethod
from typing import Any, Callable, Iterable, Sequence, Union

import numpy as np

LikeProperty = Union["Property", float, int, str]
PropertySequence = Union[Sequence[LikeProperty], np.ndarray]


class Operation(ABC):
    def __call__(self, **params) -> Callable:
        def inner(obj):
            self.operation(obj, **params)

        return inner

    @abstractmethod
    def operation(self, obj, **params) -> Any:
        return NotImplemented


class Property(ABC):
    @abstractmethod
    def get(self) -> Any:
        return NotImplemented

    @property
    def value(self) -> Any:
        return self.get()

    def transform(self, operation: Operation, **kwargs) -> "Property":
        assert isinstance(operation, Operation)
        return operation(**kwargs)(self)


class Constant(Property):
    def __init__(self, value):
        self._value = value

    def get(self) -> Any:
        return self._value


def as_property(value: LikeProperty) -> Property:
    if isinstance(value, Property):
        return value
    return Constant(value)


class Source(Property):
    def __init__(self, fn: Callable, *args, **kwargs):
        assert callable(fn)
        self._fn = fn
        self._args = args
        self._kwargs = kwargs

    def get(self) -> Any:
        return self._fn(*self._args, **self._kwargs)


class Convert(Source):
    def __init__(self, p: LikeProperty, fn: Callable, *args, **kwargs):
        self._original = as_property(p)
        super().__init__(fn, *args, **kwargs)

    def get(self) -> Any:
        return self._fn(self._original.value, *self._args, **self._kwargs)


class Iter(Property):
    def __init__(self, seq: Iterable):
        self._source = seq
        self._iter = iter(self._source)

    def _next(self):
        try:
            return next(self._iter)
        except StopIteration:
            self._iter = iter(self._source)
            return next(self._iter)

    def get(self) -> Any:
        return self._next()


# TODO: random state
class RandomUniform(Property):
    def __init__(self, low=0.0, high=1.0):
        self._low = low
        self._high = high

    def get(self):
        return np.random.uniform(self._low, self._high)


# TODO: random state
class RandomChoice(Property):
    def __init__(self, items):
        self._items = list(items)

    def get(self):
        return np.random.choice(self._items)


class Transformer:
    def __init__(self, *operations: Operation):
        assert all([isinstance(op, Operation) for op in operations])
        self._operations = operations

    def __call__(self, instance: Property) -> Property:
        for op in self._operations:
            instance = instance.transform(op)
        return instance



import numpy as np
from pyece.core.math.rotate import rotate


def test_rotate_2d():
    assert np.allclose(
        rotate((1, 1), (1, 11), (np.pi / 2,)),
        np.array((-9,  1)),
        atol=1e-06
    )
    assert np.allclose(
        rotate((1, 1), (1, 11), (-np.pi / 2,)),
        np.array((11,  1)),
        atol=1e-06
    )
    assert np.allclose(
        rotate((1, 1), (1, 11), (np.pi,)),
        np.array((1,  -9)),
        atol=1e-06
    )
    assert np.allclose(
        rotate((1, 1), (1, 11), (np.pi / 6,)),
        np.array((1 - 10 / 2,  1 + 10 * np.sqrt(3) / 2)),
        atol=1e-06
    )


def test_rotate_3d():
    assert np.allclose(
        rotate((1, 1, 2), (1, 11, 2), (0, 0, np.pi / 2)),
        np.array((-9, 1, 2)),
        atol=1e-06
    )
    assert np.allclose(
        rotate((1, 2, 1), (11, 2, 1), (0, np.pi / 2, 0)),
        np.array((1, 2, -9)),
        atol=1e-06
    )
    assert np.allclose(
        rotate((2, 1, 1), (2, 1, 11), (np.pi / 2, 0, 0)),
        np.array((2, -9, 1)),
        atol=1e-06
    )



__all__ = ("rotate",)

from typing import Tuple, Union

import numpy as np


def rotate(
    pivot: np.ndarray,
    turning: np.ndarray,
    angle: Union[Tuple[float], Tuple[float, float, float]],
) -> np.ndarray:
    pivot = np.asarray(pivot)
    turning = np.asarray(turning)
    assert pivot.shape == turning.shape
    matrix = get_rotate_matrix(angle)
    return matrix @ (turning - pivot) + pivot


def get_rotate_matrix(
    angle: Union[Tuple[float], Tuple[float, float, float]]
) -> np.ndarray:
    if len(angle) == 1:
        return rotate_matrix_2d(angle[0])
    elif len(angle) == 3:
        return rotate_matrix_3d(angle)
    else:
        raise ValueError(f"ndim should be 2 or 3")


def rotate_matrix_2d(angle: float) -> np.ndarray:
    return np.array([[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]])


def rotate_matrix_3d(angle: Tuple[float, float, float]) -> np.ndarray:
    rotate_matrix_x = np.array(
        [
            [1, 0, 0],
            [0, np.cos(angle[0]), -np.sin(angle[0])],
            [0, np.sin(angle[0]), np.cos(angle[0])],
        ]
    )
    rotate_matrix_y = np.array(
        [
            [np.cos(angle[1]), 0, np.sin(angle[1])],
            [0, 1, 0],
            [-np.sin(angle[1]), 0, np.cos(angle[1])],
        ]
    )
    rotate_matrix_z = np.array(
        [
            [np.cos(angle[2]), -np.sin(angle[2]), 0],
            [np.sin(angle[2]), np.cos(angle[2]), 0],
            [0, 0, 1],
        ]
    )
    return rotate_matrix_x @ rotate_matrix_y @ rotate_matrix_z



import typing as tp

import h5py as h5  # type: ignore
import numpy as np
from .core.property import Corners

import itertools


def linspace(start, end, count: int):
    # NOTE: np.linspace works ~ 8 times faster
    start = np.asarray(start)
    end = np.asarray(end)
    C = (end - start) / (count)
    return start + np.asarray([C * (k + 0.5) for k in range(count)])


# TODO: do it asynchronously
def meshcorners(corners: np.ndarray, grid: tp.Tuple):
    c = len(corners) // 2
    if c == 0:
        return corners[0]

    return linspace(
        meshcorners(corners[:c], grid[1:]), meshcorners(corners[c:], grid[1:]), grid[0]
    )


def cutpatch(
    data: tp.Union[np.ndarray, h5.Dataset],
    corners: np.ndarray,
    grid: tp.Tuple[int, ...],
    fill: tp.Any = None,
) -> np.ndarray:
    mesh = np.asarray(meshcorners(corners, grid)).round()
    idx = np.rollaxis(mesh.astype(int), -1)
    d = idx.shape[0]
    ring_idx = idx % np.reshape(data.shape[:d], (d, *[1] * d))
    if isinstance(data, h5.Dataset):
        data = data[:]
    patch = data[tuple(ring_idx)]
    if fill is not None:
        fill_idx = (idx != ring_idx).any(0)
        patch[fill_idx] = fill
    return patch


class ByPatchWrapper:
    def __init__(
        self,
        func: tp.Callable[[np.ndarray], tp.Any],
        size: tp.Union[np.ndarray, tp.Tuple[int, ...]],
        spacing: tp.Union[np.ndarray, tp.Tuple[float, ...]] = None,
        grid: tp.Union[np.ndarray, tp.Tuple[float, ...]] = None,
    ):
        assert callable(func)
        self._wrapper = func
        dim = len(size)
        size = np.asarray(size)
        spacing = np.asarray(spacing or [1.0] * dim)

        if grid is not None:
            grid = np.asarray(grid).astype(int)

        assert len(spacing) == dim
        assert (grid is None) or (len(grid) == dim)

        self._dim = dim
        self._size = volume = size
        if spacing is not None:
            volume = volume * spacing
        self._volume = volume
        self._corners = Corners.product(tuple(volume)).value

        self._grid = grid

    def __call__(
        self,
        data: np.ndarray,
        spacing: tp.Union[np.ndarray, tp.Tuple[float, ...]] = None,
        grid: tp.Union[np.ndarray, tp.Tuple[float, ...]] = None,
        **kwargs,
    ) -> np.ndarray:
        size = np.asarray(data.shape[: self._dim])
        if spacing is not None:
            size = size * np.asarray(spacing)

        grid = np.asarray(
            grid or self._grid or np.floor((size / self._volume)), dtype="int"
        )
        ids = np.asarray(list(itertools.product(*map(range, grid))))
        shift = self._volume + ((size - (grid * self._volume)) / (grid - 1))
        shifts = ids * shift

        wrapped = list()
        for shift in shifts:
            corners = self._corners + shift
            if spacing is not None:
                corners = corners / spacing
            corners = corners - 0.5
            patch = cutpatch(
                data=data,
                grid=tuple(self._size),
                corners=corners,
            ).astype(np.float32)

            wrapped_patch = self._wrapper(patch, **kwargs)
            wrapped.append(wrapped_patch)

        result = np.asarray(wrapped)
        tail_shape = result.shape[1:]
        return result.reshape((*grid, *tail_shape))



import numpy as np
from typing import List

class Point:
    def __init__(self, *coords: AnyPoint) -> None: ...
    @property
    def coords(self): ...
    @coords.setter
    def coords(self, coords) -> None: ...
    def numpy(self) -> np.ndarray: ...
    def __len__(self): ...
    def __getitem__(self, i) -> Point: ...
    def __setitem__(self, i, v) -> None: ...
    def __iter__(self): ...
    def __eq__(self, other: Point) -> bool: ...
    def __hash__(self): ...
    def __add__(self, other): ...
    def __sub__(self, other): ...
    def __mul__(self, other): ...
    def __truediv__(self, other): ...
    def __radd__(self, other): ...
    def __rsub__(self, other): ...
    def __rmul__(self, other): ...
    def __rtruediv__(self, other): ...

class Size:
    def __init__(self, *sides: AnySize) -> None: ...
    @property
    def sides(self): ...
    @sides.setter
    def sides(self, other) -> None: ...
    def numpy(self) -> np.ndarray: ...
    def __len__(self): ...
    def __getitem__(self, i) -> Size: ...
    def __setitem__(self, i, v) -> None: ...
    def __iter__(self): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def __hash__(self): ...

class Box:
    def __init__(self, anchor: Point, sides: Size, canvas: Size) -> None: ...
    def __len__(self): ...
    @property
    def anchor(self) -> Point: ...
    @anchor.setter
    def anchor(self, other) -> None: ...
    @property
    def sides(self) -> Size: ...
    @sides.setter
    def sides(self, other) -> None: ...
    @property
    def distant(self) -> Point: ...
    @property
    def centre(self) -> Point: ...
    @property
    def canvas(self) -> Size: ...
    @canvas.setter
    def canvas(self, other) -> None: ...
    def __hash__(self): ...
    def __contains__(self, point: Point): ...
    def __eq__(self, other: Box) -> bool: ...
    def __ne__(self, other: Box) -> bool: ...
    @property
    def area(self) -> float: ...
    def split(self, *points: Point) -> List['Box']: ...

def area_union(*boxes: Box) -> float: ...
def area_intersection(*box_groups: List[Box]) -> float: ...



__all__ = [
    'Point',
    'Size',
    'Box',
    'area_union',
    'area_intersection',
]

from collections.abc import Iterable
from typing import Dict, List, Tuple, Union

import numpy as np

Number = (int, float, complex)
AnyPoint = Union[Tuple[Union[int, float, complex], ...], 'Point']
AnySize = Union[Tuple[Union[int, float, complex], ...], 'Size']

def _as_numpy(fn):
    def wrapper(self, other):
        return Point(fn(self.numpy(), Point(other).numpy()))
    return wrapper


class Point:
    def __init__(self, *coords: AnyPoint):
        if len(coords) == 1:
            coords = coords[0]
            if np.isscalar(coords):
                coords = [coords]
            elif isinstance(coords, Point):
                coords = coords
            elif isinstance(coords, Iterable):
                coords = Point(*coords)
            else:
                raise NotImplementedError
        self.coords = coords

    @property
    def coords(self):
        return self._coords

    @coords.setter
    def coords(self, coords):
        if isinstance(coords, Point):
            self._coords = coords.coords
        self._coords = [
            p if np.isscalar(p) else Point(p)
            for p in coords
        ]

    def numpy(self) -> np.ndarray:
        return np.asarray(self.coords, dtype=float)

    def __len__(self):
        return len(self._coords)

    def __getitem__(self, i) -> 'Point':
        return self._coords[i]

    def __setitem__(self, i, v) -> None:
        if all([np.isscalar(v) for v in self]):
            self._coords[i]=v
            return
        raise NotImplementedError

    def __iter__(self):
        for s in self.coords:
            yield s

    def __repr__(self) -> str:
        return 'Point[{}]'.format(', '.join([*map(repr, self.coords)]))

    def __eq__(self, other: 'Point') -> bool:
        return hash(self) == hash(Point(other))

    def __hash__(self):
        return hash(self.numpy().data.tobytes())

    @_as_numpy
    def __add__(self, other):
        return self + other

    @_as_numpy
    def __sub__(self, other):
        return self - other

    @_as_numpy
    def __mul__(self, other):
        return self * other

    @_as_numpy
    def __truediv__(self, other):
        return self / other

    @_as_numpy
    def __radd__(self, other):
        return self + other

    @_as_numpy
    def __rsub__(self, other):
        return self - other

    @_as_numpy
    def __rmul__(self, other):
        return self * other

    @_as_numpy
    def __rtruediv__(self, other):
        return self / other

class Size:
    def __init__(self, *sides: AnySize):
        if len(sides) == 1:
            sides = sides[0]
            if np.isscalar(sides):
                sides = [sides]
            elif isinstance(sides, Size):
                sides = sides
            elif isinstance(sides, Iterable):
                sides = Size(*sides)
            else:
                raise NotImplementedError
        self.sides = sides

    @property
    def sides(self):
        return self._sides

    @sides.setter
    def sides(self, other):
        if all([np.isscalar(s) for s in other]):
            self._sides = [s for s in other]
        else:
            raise NotImplementedError

    def numpy(self) -> np.ndarray:
        return np.asarray(self.sides, dtype=float)

    def __len__(self):
        return len(self._sides)

    def __getitem__(self, i) -> 'Size':
        return self._sides[i]

    def __setitem__(self, i, v) -> None:
        if all([np.isscalar(v) for v in self]):
            self._sides[i] = v
            return
        raise NotImplementedError

    def __iter__(self):
        for s in self.sides:
            yield s

    def __eq__(self, other):
        return hash(self) == hash(Size(other))

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return 'Size[{}]'.format(', '.join([*map(repr, self.sides)]))

    def __hash__(self):
        return hash(self.numpy().data.tobytes())


class Box:
    def __init__(self, anchor: Point, sides: Size, canvas: Size):
        self._dim = None
        self.canvas = canvas
        self.anchor = anchor
        self.sides = sides

    def __len__(self):
        return self._dim

    @property
    def anchor(self) -> Point:
        return Point(self._a) * self.canvas

    @anchor.setter
    def anchor(self, other):
        anchor = Point(other) / self.canvas
        assert len(anchor) == len(self)
        self._a = anchor

    @property
    def sides(self) -> Size:
        return Size(Point(self._s) * self.canvas)

    @sides.setter
    def sides(self, other) -> None:
        sides = Point(other)
        assert len(sides) == len(self)
        for dim, s in enumerate(sides):
            if s < 0:
                a = self.anchor
                a[dim] += s
                sides[dim] = -1 * s
                self.anchor = a
        sides = Size(sides / self.canvas)
        self._s = sides

    @property
    def distant(self) -> Point:
        return self.anchor + self.sides

    @property
    def centre(self) -> Point:
        return (self.anchor + self.distant) / 2

    @property
    def canvas(self) -> Size:
        return self._c

    @canvas.setter
    def canvas(self, other) -> None:
        canvas = Size(other)
        if self._dim is None:
            self._dim = len(canvas)
        else:
            assert len(canvas) == len(self)
        self._c = canvas

    def __repr__(self):
        return 'Box[{}+>{}]'.format(repr(self.anchor.coords), repr(self.sides.sides))

    def __hash__(self):
        return hash(
            self._a.numpy().data.tobytes() +
            self._s.numpy().data.tobytes()
        )

    def __contains__(self, point: Point):
        point = Point(point)
        p1 = self.anchor
        p2 = p1 + self.sides
        return all([p1[dim] < point[dim] < p2[dim] for dim in range(len(self))])

    def __eq__(self, other: 'Box') -> bool:
        if isinstance(other, Box):
            return hash(self) == hash(other)
        raise NotImplementedError

    def __ne__(self, other: 'Box') -> bool:
        return not self == other

    @property
    def area(self) -> float:
        return np.prod(Point(self.sides))

    def split(self, *points: Point) -> List['Box']:
        ps = [Point(p) for p in points]
        assert all([len(p) == len(self) for p in ps])
        p1 = self.anchor
        p2 = p1 + self.sides
        ax = [
            [p[dim] for p in ps if p1[dim] < p[dim] < p2[dim]] + [p1[dim]] + [p2[dim]]
            for dim in range(len(self))
        ]
        for dim, a in enumerate(ax):
            a.sort()
            ax[dim] = [p for p in zip(a[:-1], a[1:])]
        idxs = np.meshgrid(*[range(len(a)) for a in ax], indexing='ij')
        idxs = np.asarray([idx.flatten() for idx in idxs]).transpose()
        result = list()
        canvas = self.canvas
        for idx in idxs:
            p1 = Point([ax[dim][i][0] for dim, i in enumerate(idx)])
            p2 = Point([ax[dim][i][1] for dim, i in enumerate(idx)])
            box = Box(p1, p2-p1, canvas)
            result.append(box)
        return result

def area_union(*boxes: Box) -> float:
    assert len(set(box.canvas for box in boxes)) == 1
    split_points = set(
        [box.anchor for box in boxes] +
        [box.distant for box in boxes]
    )
    unique_boxs = set()
    for box in boxes:
        unique_boxs.update(box.split(*split_points))
    return np.sum([box.area for box in unique_boxs])

def map_box_groups(*box_groups: List[Box]) -> Dict[int, List[Box]]:
    groups = {i: list() for i in range(len(box_groups))}
    for i, group in enumerate(box_groups):
        if isinstance(group, Iterable):
            for box in group:
                assert isinstance(box, Box)
                groups[i].append(box)
        elif isinstance(group, Box):
            groups[i].append(group)
        else:
            raise NotImplementedError
    assert len(set(box.canvas for group in groups.values() for box in group))
    return groups

def area_intersection(*box_groups: List[Box]) -> float:
    groups = map_box_groups(*box_groups)
    corners = {
        p for group in groups.values()
        for box in group
        for p in [box.anchor, box.distant]}
    groups = {i: {b for box in group for b in box.split(*corners)} for i, group in groups.items()}
    intersection = set.intersection(*list(groups.values()))
    return np.sum([box.area for box in intersection])



from pytest import mark
import pytest
from pyece import Corners


@mark.parametrize(
    "shape",
    [(1, 2), (1, 2, 3), (1, 2, 3, 4)],
)
def test_corners_product(shape):
    cs = Corners.product(shape).value
    assert cs.shape[0] == 2 ** len(shape)
    assert cs.mean() == sum(shape) / len(shape) / 2
    assert (cs[0] == [0]*len(shape)).all()
    assert (cs[-1] == shape).all()
    assert set(cs.flatten()) == {0, *shape}


def test_corners():
    assert Corners([[]]).value.shape == (1, 0)
    with pytest.raises(ValueError):
        Corners([[1]]).value
    with pytest.raises(ValueError):
        Corners([[0, 0]]*3).value
    assert Corners([[0, 0]]*4).value.shape == (4,2)
    with pytest.raises(ValueError):
        Corners([[0, 0]]*8).value
    assert Corners([[0]*3]*8).value.shape == (8, 3)
    assert Corners([[0]*4]*16).value.shape == (16, 4)



__all__ = (
    "Point",
    "PointOperation",
    "PointShift",
    "PointRotate",
    "PointInflation",
)

from abc import abstractmethod
from typing import Callable, Union

import numpy as np

from ..math.rotate import rotate
from .base import LikeProperty, Operation, Property, PropertySequence, as_property


class Point(Property):
    def __init__(self, point: PropertySequence):
        self._point = [as_property(p) for p in point]

    def get(self) -> np.ndarray:
        return np.asarray([p.value for p in self._point])


LikePoint = Union[Point, PropertySequence, np.ndarray]


def as_point(value: LikePoint) -> Point:
    if isinstance(value, Point):
        return value
    return Point(value)


class PointOperation(Operation):
    def __call__(self, **params) -> Callable[[LikePoint], Point]:
        def inner(point: LikePoint):
            point = as_point(point).value
            value = self.operation(point, **params)
            return Point(value)

        return inner

    @abstractmethod
    def operation(self, obj: np.ndarray, **params) -> np.ndarray:
        return NotImplemented


class PointShift(PointOperation):
    def __init__(self, shift: LikePoint):
        self._shift = as_point(shift)

    def __call__(self) -> Callable[[LikePoint], Point]:
        shift = self._shift.value
        return super().__call__(shift=shift)

    def operation(self, obj: np.ndarray, shift: np.ndarray) -> np.ndarray:
        return obj + shift


class PointRotate(PointOperation):
    def __init__(self, angle: LikeProperty, pivot: LikePoint = None):
        self._angle = as_property(angle)
        self._pivot = None if pivot is None else as_point(pivot)

    def __call__(self, pivot: LikePoint = None) -> Callable[[LikePoint], Point]:
        angle = np.asarray(self._angle.value).reshape(-1) % (2 * np.pi)
        if pivot is not None:
            pivot = as_point(pivot).value
        elif self._pivot is not None:
            pivot = self._pivot.value
        else:
            raise RuntimeError
        return super().__call__(pivot=pivot, angle=angle)

    def operation(
        self, obj: np.ndarray, pivot: np.ndarray, angle: np.ndarray
    ) -> np.ndarray:
        return rotate(pivot, obj, angle)


class PointInflation(PointOperation):
    def __init__(self, factor: LikeProperty, pivot: LikePoint = None):
        self._factor = as_property(factor)
        self._pivot = None if pivot is None else as_point(pivot)

    def __call__(self, pivot: LikePoint = None) -> Callable[[LikePoint], Point]:
        factor = self._factor.value
        if pivot is not None:
            pivot = as_point(pivot).value
        elif self._pivot is not None:
            pivot = self._pivot.value
        else:
            raise RuntimeError
        return super().__call__(pivot=pivot, factor=factor)

    def operation(
        self, obj: np.ndarray, pivot: np.ndarray, factor: np.ndarray
    ) -> np.ndarray:
        return pivot + (obj - pivot) * factor



__all__ = (
    "PointCloud",
    "Corners",
)

from itertools import product
from typing import Sequence, Tuple, Union

import numpy as np

from .base import Operation, Property
from .point import (
    LikePoint,
    Point,
    PointInflation,
    PointOperation,
    PointRotate,
    as_point,
)


class PointCloud(Point):
    def __init__(self, points: Sequence[LikePoint]):
        super().__init__([as_point(p) for p in points])

    def transform(self, operation: Operation, **kwargs) -> Property:
        if isinstance(operation, PointOperation):
            points = self.value
            if isinstance(operation, (PointRotate, PointInflation)):
                if operation._pivot is None:
                    kwargs["pivot"] = points.mean(0)
            fn = operation(**kwargs)
            return PointCloud(tuple(map(fn, points)))
        return super().transform(operation, **kwargs)


LikePointCloud = Union[PointCloud, Sequence[LikePoint]]


def as_pointcloud(value: LikePointCloud) -> PointCloud:
    if isinstance(value, PointCloud):
        return value
    return PointCloud(value)


class Corners(PointCloud):
    def get(self) -> np.ndarray:
        corners = super().get()
        n, p = corners.shape
        if 2 ** p != n:
            msg = ""
            raise ValueError(msg)
        return corners

    @staticmethod
    def product(shape: Tuple) -> "Corners":
        masks = list(product((False, True), repeat=len(shape)))
        corners = [[s if c else 0 for c, s in zip(mask, shape)] for mask in masks]
        return Corners(corners)



import h5py as h5
import numpy as np
import typing as tp
from .core.property import Corners as Corners

def linspace(start, end, count: int): ...
def meshcorners(corners: np.ndarray, grid: tp.Tuple): ...
def cutpatch(data: tp.Union[np.ndarray, h5.Dataset], corners: np.ndarray, grid: tp.Tuple[int, ...], fill: tp.Any = ...) -> np.ndarray: ...

class ByPatchWrapper:
    def __init__(self, func: tp.Callable[[np.ndarray], tp.Any], size: tp.Union[np.ndarray, tp.Tuple[int, ...]], spacing: tp.Union[np.ndarray, tp.Tuple[float, ...]] = ..., grid: tp.Union[np.ndarray, tp.Tuple[float, ...]] = ...) -> None: ...
    def __call__(self, data: np.ndarray, spacing: tp.Union[np.ndarray, tp.Tuple[float, ...]] = ..., grid: tp.Union[np.ndarray, tp.Tuple[float, ...]] = ..., **kwargs) -> np.ndarray: ...



import abc
import numpy as np
from abc import ABC, abstractmethod
from typing import Any, Callable, Iterable, Sequence, Union

PropertySequence = Union[Sequence[LikeProperty], np.ndarray]

class Operation(ABC, metaclass=abc.ABCMeta):
    def __call__(self, **params) -> Callable: ...
    @abstractmethod
    def operation(self, obj, **params) -> Any: ...

class Property(ABC, metaclass=abc.ABCMeta):
    @abstractmethod
    def get(self) -> Any: ...
    @property
    def value(self) -> Any: ...
    def transform(self, operation: Operation, **kwargs) -> Property: ...

class Constant(Property):
    def __init__(self, value) -> None: ...
    def get(self) -> Any: ...

class Source(Property):
    def __init__(self, fn: Callable, *args, **kwargs) -> None: ...
    def get(self) -> Any: ...

class Convert(Source):
    def __init__(self, p: LikeProperty, fn: Callable, *args, **kwargs) -> None: ...
    def get(self) -> Any: ...

class Iter(Property):
    def __init__(self, seq: Iterable) -> None: ...
    def get(self) -> Any: ...

class RandomUniform(Property):
    def __init__(self, low: float = ..., high: float = ...) -> None: ...
    def get(self): ...

class RandomChoice(Property):
    def __init__(self, items) -> None: ...
    def get(self): ...

class Transformer:
    def __init__(self, *operations: Operation) -> None: ...
    def __call__(self, instance: Property) -> Property: ...



import abc
import numpy as np
from .base import LikeProperty, Operation, Property, PropertySequence
from abc import abstractmethod
from typing import Callable, Union

class Point(Property):
    def __init__(self, point: PropertySequence) -> None: ...
    def get(self) -> np.ndarray: ...
LikePoint = Union[Point, PropertySequence, np.ndarray]

class PointOperation(Operation, metaclass=abc.ABCMeta):
    def __call__(self, **params) -> Callable[[LikePoint], Point]: ...
    @abstractmethod
    def operation(self, obj: np.ndarray, **params) -> np.ndarray: ...

class PointShift(PointOperation):
    def __init__(self, shift: LikePoint) -> None: ...
    def __call__(self) -> Callable[[LikePoint], Point]: ...
    def operation(self, obj: np.ndarray, shift: np.ndarray) -> np.ndarray: ...

class PointRotate(PointOperation):
    def __init__(self, angle: LikeProperty, pivot: LikePoint = ...) -> None: ...
    def __call__(self, pivot: LikePoint = ...) -> Callable[[LikePoint], Point]: ...
    def operation(self, obj: np.ndarray, pivot: np.ndarray, angle: np.ndarray) -> np.ndarray: ...

class PointInflation(PointOperation):
    def __init__(self, factor: LikeProperty, pivot: LikePoint = ...) -> None: ...
    def __call__(self, pivot: LikePoint = ...) -> Callable[[LikePoint], Point]: ...
    def operation(self, obj: np.ndarray, pivot: np.ndarray, factor: np.ndarray) -> np.ndarray: ...
